import java.text.SimpleDateFormat
import java.util.Calendar
import org.apache.commons.lang.StringUtils
import org.apache.log4j.Logger
import com.google.common.collect.ArrayListMultimap
import com.atlassian.jira.issue.Issue


ganttChart = new GanttChartTool(issues, startDateCustomFieldId, endDateCustomFieldId, progressPercentageCustomFieldId, viewPeriod)

/**
 * Velocity tool to help rendering Gantt charts with the JIRA PDF View Plugin.
 * @see gantt-chart-fo.vm as the actual template.
 */
public class GanttChartTool {
	def log = Logger.getLogger(this.getClass())

	def startDateCustomFieldName
	def endDateCustomFieldName
	def progressPercentageCustomFieldName

	/**
	 * The dates to iterate over, defined by the span of the tasks.
	 */
	def dates = new ArrayList()
	def earliestDate = null
	def latestDate = null

	def earliestDateDayOfWeek = -1
	def currentDateIndex = -1
	long millisPerUnitPeriod

	def hasSubTasks
	def subTasksByParentId = ArrayListMultimap.create()

	def static calendar = Calendar.instance
	def static monthOnlyDateFormat = new SimpleDateFormat("MMM")
	def static dayOnlyDateFormat = new SimpleDateFormat("d")

	static long MILLIS_PER_DAY = 24*60*60*1000
	static long MILLIS_PER_WEEK = 7*MILLIS_PER_DAY

	public GanttChartTool(issues, startDateCustomFieldId, endDateCustomFieldId, progressPercentageCustomFieldId, viewPeriod) {
		startDateCustomFieldName = "customfield_${startDateCustomFieldId}"
		endDateCustomFieldName = "customfield_${endDateCustomFieldId}"
		  progressPercentageCustomFieldName = "customfield_${progressPercentageCustomFieldId}"

		calendar.clear()

		// find the earliest and latest dates for fully defined tasks
		for(issue in issues) {
			def startDate = getIssueStartDate(issue)
			def endDate = getIssueEndDate(issue)
			if((startDate != null) && (endDate != null)) {
				if(startDate != null) {
					if((earliestDate == null) || startDate.before(earliestDate)) {
						earliestDate = startDate
					}
				}

				if(endDate != null) {
					if((latestDate  == null) || endDate.after(latestDate)) {
						latestDate = endDate
					}
				}
			}

			  // build issue to subtasks mapping
			if(issue.isSubTask()) {
				hasSubTasks = true;
				  subTasksByParentId.put(issue.parentId, issue)
			}
		}

		// generate the dates spanned
		millisPerUnitPeriod = (viewPeriod == "day") ? MILLIS_PER_DAY : MILLIS_PER_WEEK
		if((earliestDate != null) && (latestDate != null)) {
			earliestDate = trimToDay(earliestDate, Calendar.MONDAY)
			latestDate = trimToDay(new Date(latestDate.time + MILLIS_PER_WEEK), Calendar.SUNDAY)

			Date date = earliestDate
			def i = 0
			while(!date.after(latestDate)) {
				dates.add(date)
				i += 1
				date = new Date(earliestDate.time + i*millisPerUnitPeriod)
			}

			calendar.setTime(earliestDate)
			earliestDateDayOfWeek = calendar.get(Calendar.DAY_OF_WEEK)

			def currentDate = trimToMidnight(new Date())
			currentDateIndex = (int)((currentDate.time - earliestDate.time) / millisPerUnitPeriod)
		}
	}

	/**
	 * Returns the length of the time period spanned by the issues in unit periods, or -1 if undefined.
	 */
	public int getLength() {
		if((earliestDate == null) || (latestDate == null)) {
			return -1
		}

		long length = 1 + ((latestDate.time - earliestDate.time) / millisPerUnitPeriod)
		return length
	}

	public boolean hasSubTasks() {
		return hasSubTasks
	}

	public List getSubTasksByIssue(issue) {
		  return subTasksByParentId.get(issue.id)
	}

	public Date getIssueStartDate(Issue issue) {
		return issue.getCustomFieldValue(startDateCustomFieldName)
	}

	public Date getIssueEndDate(Issue issue) {
		return issue.getCustomFieldValue(endDateCustomFieldName)
	}

	/**
	 * Returns the 0-based index of the start date in the range of all spanned dates, or -1 if undefined.
	 */
	public int getIssueStartDateIndex(Issue issue) {
		def startDate = getIssueStartDate(issue)
		long index = (startDate != null) ? ((startDate.time - earliestDate.time) / millisPerUnitPeriod) : -1
		return index
	}

	/**
	 * @see #getIssueStartDateIndex(Task)
	 */
	public int getIssueEndDateIndex(Issue issue) {
		def endDate = getIssueEndDate(issue)
		long index = (endDate != null) ? ((endDate.time - earliestDate.time) / millisPerUnitPeriod) : -1
		return index
	}

	/**
	 * Returns the length of the issue in unit periods, or -1 if undefined.
	 */
	public int getIssueLength(Issue issue) {
		def startIndex = getIssueStartDateIndex(issue)
		def endIndex = getIssueEndDateIndex(issue)
		long length = ((startIndex != -1) && (endIndex != -1)) ? (1 + (endIndex - startIndex)) : -1
		return length
	}

	/**
	 * Returns the number of unit periods beginning from the issue end-date until the latest date, or -1 if undefined.
	 */
	public int getIssueDateFiller(Issue issue) {
		long filler = getLength() - getIssueEndDateIndex(issue) - 1
		return filler
	}

	public int getIssueProgressPercentage(Issue issue) {
		  def progressPercentage = issue.getCustomFieldValue(progressPercentageCustomFieldName)
		  return (progressPercentage != null) ? progressPercentage : 0
	}

	/**
	 * Change this method if you have a different definition of "done".
	 */
	public boolean isIssueDone(Issue issue) {
		  return issue.statusObject.name in ["Closed"] // return (issue.resolutionDate != null)
	}

	public static String getMonthByDate(date) {
		return monthOnlyDateFormat.format(date)
	}

	public static String getDayOfMonthByDate(date) {
		return dayOnlyDateFormat.format(date)
	}

	/**
	 * Returns the midnight moment of the given date.
	 */
	private static trimToMidnight(date) {
		  return trimToDay(date, null)
	}

	/**
	 * Returns the midnight moment in the given day of the given date.
	 */
	private static trimToDay(date, dayOfWeek) {
		calendar.setTime(date)
		if(dayOfWeek != null) {
			  calendar.set(Calendar.DAY_OF_WEEK, dayOfWeek)
		  }
		  calendar.set(Calendar.HOUR_OF_DAY, 0)
		  calendar.set(Calendar.MINUTE, 0)
		  calendar.set(Calendar.SECOND, 0)
		  calendar.set(Calendar.MILLISECOND, 0)
		  return calendar.time
	}
}
